import importlib
import logging
import os
import signal
import sys
import time
from io import BytesIO

import gevent
import gevent.monkey
import gevent.pool
import gevent.server

from scotchwsgi import const
from scotchwsgi.response import WSGIResponseWriter
from scotchwsgi.request import WSGIRequest

logger = logging.getLogger(__name__)

class WSGIWorker(object):
    def __init__(self, app_location, sock, hostname, parent_pid):
        gevent.monkey.patch_all()

        # Ignore interrupts to disable KeyboardInterrupt being logged
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        # Allow app location to refer to files in cwd
        sys.path.append(os.getcwd())

        self.sock = sock
        self.hostname = hostname
        _, self.port = sock.getsockname()
        self.parent_pid = parent_pid

        app_module = importlib.import_module(app_location)
        if not hasattr(app_module, 'app'):
            logger.error("'app' not found in {}".format(app_location))
            sys.exit(1)

        self.application = app_module.app

    def start(self):
        logger.info("Worker starting (PID: %d)", os.getpid())

        pool = gevent.pool.Pool(size=const.MAX_CONNECTIONS)

        server = gevent.server.StreamServer(
            self.sock,
            self._handle_connection,
            spawn=pool,
        )
        server.start()

        while True:
            if os.getppid() != self.parent_pid:
                logger.info("Worker parent changed, exiting")
                break
            time.sleep(1)

    def _handle_connection(self, conn, addr):
        logger.info("New connection: %s", addr)

        reader = conn.makefile('rb')
        writer = conn.makefile('wb')

        try:
            request = WSGIRequest.from_reader(reader)
        except ValueError:
            # TODO: Send invalid request response
            logger.error("Invalid request received from: %s", addr)
            pass
        else:
            self._send_response(request, writer)

        logger.debug("Closing connection")

        try:
            reader.close()
        except IOError:
            pass

        try:
            writer.close()
        except IOError:
            pass

        conn.close()

    def _send_response(self, request, writer):
        environ = self._get_environ(request)
        response_writer = WSGIResponseWriter(writer)

        logger.debug("Calling into application")
        response_iter = self.application(environ, response_writer.start_response)

        try:
            for response in response_iter:
                logger.debug("Write %s", response)
                response_writer.write(response)
        except Exception as e:
            logger.error("Application aborted: %r", e)
        finally:
            response_iter_close = getattr(response_iter, 'close', None)
            if callable(response_iter_close):
                response_iter.close()
            logger.debug("Called into application")

    def _get_environ(self, request):
        environ = {
            'REQUEST_METHOD': request.method,
            'SCRIPT_NAME': '',
            'SERVER_NAME': self.hostname,
            'SERVER_PORT': str(self.port),
            'SERVER_PROTOCOL': request.http_version,
            'PATH_INFO': request.path,
            'QUERY_STRING': request.query,
            'wsgi.version': (1, 0),
            'wsgi.url_scheme': 'http',
            'wsgi.input': BytesIO(request.body),
            'wsgi.errors': sys.stderr,
            'wsgi.multithread': True,
            'wsgi.multiprocess': True,
            'wsgi.run_once': False,
        }

        environ['PATH_INFO'] = request.path
        environ['QUERY_STRING'] = request.query if request.query else ''

        environ_headers = request.headers.copy()
        if 'content-type' in request.headers:
            environ['CONTENT_TYPE'] = environ_headers.pop('content-type')
        if 'content-length' in request.headers:
            environ['CONTENT_LENGTH'] = environ_headers.pop('content-length')
        for http_header_name, http_header_value in environ_headers.items():
            http_header_name = 'HTTP_{}'.format(http_header_name.upper().replace('-', '_'))
            environ[http_header_name] = http_header_value
        environ_headers.clear()

        return environ

def start_new_worker(*args, **kwargs):
    worker = WSGIWorker(*args, **kwargs)
    worker.start()
    return worker



import logging

from scotchwsgi import const

logger = logging.getLogger(__name__)

class WSGIRequest(object):
    def __init__(self, method, path, query, http_version, headers, body):
        self.method = method
        self.path = path
        self.query = query
        self.http_version = http_version
        self.headers = headers
        self.body = body

    @staticmethod
    def read_request_line(reader):
        request_line = reader.readline().decode(const.STR_ENCODING)
        logger.info("Received request %s", request_line)

        try:
            request_method, request_uri, http_version = request_line.split(' ', 3)
        except ValueError:
            raise ValueError("Invalid request line: %s" % request_line)

        http_version = http_version.rstrip()
        request_uri_split = request_uri.split('?', 1)
        request_path = request_uri_split[0]
        if len(request_uri_split) > 1:
            request_query = request_uri_split[1]
        else:
            request_query = ''

        return request_method, request_path, request_query, http_version

    @staticmethod
    def read_headers(reader):
        headers = {}
        while True:
            header = reader.readline().decode(const.STR_ENCODING).replace('\r\n', '\n').rstrip('\n')
            if header == '':
                break

            try:
                header_name, header_value = header.split(':', 1)
            except ValueError:
                raise ValueError("Invalid header: %s" % header)

            header_name = header_name.lower()
            header_value = header_value.lstrip()
            headers[header_name] = header_value

        logger.debug("Headers: %s", headers)

        return headers

    @staticmethod
    def read_body(reader, content_length):
        if content_length > 0:
            logger.debug("Reading body")
            message_body = reader.read(content_length)
            logger.debug("Body: %s", message_body)
        else:
            logger.debug("No body")
            message_body = b""

        if content_length != len(message_body):
            raise ValueError(
                "content-length %d too large, only read %d bytes" % (
                    content_length, len(message_body)
                )
            )

        return message_body

    @staticmethod
    def from_reader(reader):
        method, path, query, http_version = WSGIRequest.read_request_line(
            reader
        )

        headers = WSGIRequest.read_headers(
            reader
        )

        body = WSGIRequest.read_body(
            reader,
            int(headers.get('content-length', 0))
        )

        return WSGIRequest(
            method=method,
            path=path,
            query=query,
            http_version=http_version,
            headers=headers,
            body=body,
        )



import logging
import multiprocessing
import os
import signal
import ssl
import sys
import time

from gevent import socket

from scotchwsgi.worker import start_new_worker

logger = logging.getLogger(__name__)

class WSGIServer(object):
    def __init__(self, host, port, app_location, ssl_config=None, backlog=None, num_workers=1):
        self.host = host
        self.port = port
        self.app_location = app_location
        self.ssl_config = ssl_config
        self.backlog = backlog
        self.num_workers = num_workers
        self.worker_processes = []

    def start(self, blocking=True):
        self.sock = socket.socket()
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.host, self.port))
        if self.ssl_config:
            logger.info("Using SSL")
            self.sock = ssl.wrap_socket(
                self.sock,
                server_side=True,
                **self.ssl_config,
            )

        if self.backlog:
            self.sock.listen(self.backlog)
        else:
            self.sock.listen()

        logger.info("Listening on %s:%d", self.host, self.port)

        for worker_index in range(self.num_workers):
            worker_process = multiprocessing.Process(
                name="worker-%d"%worker_index,
                target=start_new_worker,
                args=(
                    self.app_location,
                    self.sock,
                    self.host,
                    os.getpid(),
                )
            )
            worker_process.start()
            self.worker_processes.append(worker_process)

        signal.signal(signal.SIGTERM, self.handle_signal)
        signal.signal(signal.SIGINT, self.handle_signal)

        self.alive = True
        if blocking:
            while self.alive:
                time.sleep(1)

    def stop(self):
        for index, worker_process in enumerate(self.worker_processes):
            logger.info("Terminating worker process %d (PID: %d)", index, worker_process.pid)
            worker_process.terminate()
        self.worker_processes = []
        self.alive = False
        self.sock.close()

    def handle_signal(self, signo, _stack_frame):
        logger.debug("Received signal %d", signo)
        self.stop()

def make_server(*args, **kwargs):
    return WSGIServer(*args, **kwargs)



import logging

from scotchwsgi import const

logger = logging.getLogger(__name__)

class WSGIResponseWriter(object):
    def __init__(self, writer):
        self.writer = writer
        self.headers_to_send = []
        self.headers_sent = []

    def start_response(self, status, response_headers, exc_info=None):
        logger.debug("start_response %s %s %s", status, response_headers, exc_info)

        if exc_info:
            try:
                if self.headers_sent:
                    logger.debug("Reraising application exception")
                    # reraise original exception if headers already sent
                    raise exc_info[1].with_traceback(exc_info[2])
            finally:
                exc_info = None # avoid dangling circular ref
        elif self.headers_to_send:
            raise AssertionError("Headers already set")

        self.headers_to_send[:] = [status, response_headers]

        return self.write

    def write(self, data):
        if not self.headers_to_send:
            raise AssertionError("write() before start_response()")

        elif not self.headers_sent:
            status, response_headers = self.headers_to_send[:]
            logger.debug("Send headers %s %s", status, response_headers)

            self.writer.write(b"HTTP/1.1 ")
            self.writer.write(status.encode(const.STR_ENCODING))
            self.writer.write(b"\r\n")

            for header_name, header_value in response_headers:
                self.writer.write(header_name.encode(const.STR_ENCODING))
                self.writer.write(b": ")
                self.writer.write(header_value.encode(const.STR_ENCODING))
                self.writer.write(b"\r\n")

            self.writer.write(b"\r\n")

            self.headers_sent[:] = [status, response_headers]

        self.writer.write(data)
        self.writer.flush()
